package armazenamento;

import java.util.ArrayList;

import java.util.List;

import hibernate.HibernateTransaction;
import negocio.exception.LocacaoExistenteException;
import negocio.exception.LocacaoInexistenteException;

import org.hibernate.ObjectNotFoundException;
import org.hibernate.PropertyValueException;
import org.hibernate.Session;
import org.hibernate.exception.ConstraintViolationException;

import util.IteratorBidirecional;

import dados.exception.LocacaoInvalidaException;
import dados.locacao.Copia;
import dados.locacao.Locacao;
import dados.locacao.Periodo;
import dados.usuario.Cliente;

public class RepositorioLocacao {
	
	private static RepositorioLocacao instancia;
	
	static{
		RepositorioLocacao.instancia = new RepositorioLocacao();
	}
	
	private RepositorioLocacao() {}
	
	public static RepositorioLocacao getInstancia(){
		return RepositorioLocacao.instancia;
	}
	
	public void inserirLocacao(Locacao locacao) throws LocacaoInvalidaException, LocacaoExistenteException{
		if(locacao != null){
			HibernateTransaction busca = new HibernateTransaction();
			
			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();
				
				// AUTO INCREMENT!!!
				try{
					locacao.setId((Long)sessao.createQuery("select max(p.id) from dados.locacao.Locacao p").list().get(0) + 1);
				}catch(NullPointerException e){
					locacao.setId(0);
				}

				sessao.save(locacao);

				busca.closeSessionCommit();
			}catch(ConstraintViolationException e){
				busca.transactionRollBack();
				
				String retorno = e.getSQLException().getMessage();

				if(retorno.contains("key 1") || retorno.contains("key 2")){
					throw new LocacaoExistenteException();
				} 
				
			}catch(PropertyValueException e){
				busca.transactionRollBack();
				throw new LocacaoInvalidaException();
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public void removerLocacao(Copia copia) throws LocacaoInexistenteException{
		Locacao locacao = RepositorioLocacao.instancia.getLocacao(copia);
		
		HibernateTransaction busca = new HibernateTransaction();
		busca.openSessionBegin();
		Session sessao = busca.getSession();

		sessao.delete(locacao);
		
		busca.closeSessionCommit();
	}
	
	public boolean existeLocacao(Copia copia){
		boolean retorno = true;
		
		try{
			this.getLocacao(copia);
		} catch(LocacaoInexistenteException e){
			retorno = false;
		}
		
		return retorno;
	}
	
	public Locacao getLocacao(long id) throws LocacaoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		
		try{
			Session sessao = busca.openSession();
			Locacao locacao = (Locacao) sessao.get(Locacao.class, id);
			
			busca.closeSession();
			
			if(locacao != null){
				return locacao;
			}else{
				throw new LocacaoInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new LocacaoInexistenteException();
		}
	}
	
	@SuppressWarnings("unchecked")
	public Locacao getLocacao(Copia copia) throws LocacaoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		
		try{
			List<Locacao> lista = (List<Locacao>) busca.openSession().createQuery(String.format("from dados.locacao.Locacao where id_Copia=%d", copia.getCodigoCopia())).list();
			
			busca.closeSession();
			
			if(lista != null && !lista.isEmpty()){
				return lista.get(0);
			}else{
				throw new LocacaoInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new LocacaoInexistenteException();
		}
	}
	
	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Locacao> getLocacoes(Cliente cliente) throws LocacaoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		IteratorBidirecional<Locacao> iterador = null;
		
		try{
			List<Locacao> lista = (List<Locacao>) busca.openSession().createQuery(String.format("from dados.locacao.Locacao where id_Cliente=%d", cliente.getCpf())).list();
			
			if(!lista.isEmpty()){
				busca.closeSession();
				iterador = new IteratorBidirecional<Locacao>(lista.iterator());
			}else{
				busca.closeSession();
				throw new LocacaoInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new LocacaoInexistenteException();
		}

		return iterador;
	}
	
	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Locacao> getLocacoesInterceptaPeriodo(Copia copia, Periodo periodo) throws LocacaoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		IteratorBidirecional<Locacao> iterador = null;
		
		try{
			List<Locacao> lista = (List<Locacao>) busca.openSession().createQuery(String.format("from dados.locacao.Locacao where id_Copia = %d", copia.getCodigoCopia())).list();
			
			busca.closeSession();
			
			if(!lista.isEmpty()){
				
				List<Locacao> retorno = new ArrayList<Locacao>();
				
				for (Locacao locacao : lista) {
					
					if(locacao.getPeriodo().intersecta(periodo)){
						retorno.add(locacao);
					}
				}
				
				if(retorno.size() > 0){
					iterador = new IteratorBidirecional<Locacao>(retorno.iterator());
				} else {
					throw new LocacaoInexistenteException();
				}
			}else{
				throw new LocacaoInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new LocacaoInexistenteException();
		}

		return iterador;
	}
	
}
