package armazenamento;

import java.util.ArrayList;
import java.util.List;

import hibernate.HibernateTransaction;
import negocio.exception.ReservaExistenteException;
import negocio.exception.ReservaInexistenteException;

import org.hibernate.ObjectNotFoundException;
import org.hibernate.PropertyValueException;
import org.hibernate.Session;
import org.hibernate.exception.ConstraintViolationException;

import util.IteratorBidirecional;

import dados.locacao.ClienteCopiaPK;
import dados.locacao.Copia;
import dados.locacao.Periodo;
import dados.locacao.Reserva;
import dados.usuario.Cliente;

public class RepositorioReserva {

	private static RepositorioReserva instancia;
	
	static{
		RepositorioReserva.instancia = new RepositorioReserva();
	}
	
	private RepositorioReserva(){}
	
	public static RepositorioReserva getInstancia(){
		return RepositorioReserva.instancia;
	}
	
	public void inserirReserva(Reserva reserva) throws ReservaExistenteException{
		if(reserva != null){
			HibernateTransaction busca = new HibernateTransaction();
			
			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();

				// AUTO INCREMENT!!!
				try{
					reserva.setId((Long)sessao.createQuery("select max(p.id) from dados.locacao.Reserva p").list().get(0) + 1);
				}catch(NullPointerException e){
					reserva.setId(0);
				}

				sessao.save(reserva);

				busca.closeSessionCommit();
			}catch(ConstraintViolationException e){
				busca.transactionRollBack();
				
				String retorno = e.getSQLException().getMessage();

				if(retorno.contains("key 1")){
					throw new ReservaExistenteException(reserva);
				} 
				
			}catch(PropertyValueException e){
				busca.transactionRollBack();
				throw new ReservaExistenteException(reserva);
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public void removerReserva(Cliente cliente, Copia copia) throws ReservaInexistenteException{
		Reserva reserva = RepositorioReserva.instancia.getReserva(cliente, copia);
		
		HibernateTransaction busca = new HibernateTransaction();
		busca.openSessionBegin();
		Session sessao = busca.getSession();

		sessao.delete(reserva);
		
		busca.closeSessionCommit();
	}
	
	public boolean existeReserva(Cliente cliente, Copia copia){
		boolean retorno = true;
		
		try{
			this.getReserva(cliente,copia);
		} catch(ReservaInexistenteException e){
			retorno = false;
		}
		
		return retorno;
	}
	
	public Reserva getReserva(Cliente cliente, Copia copia) throws ReservaInexistenteException{
		ClienteCopiaPK chave = new ClienteCopiaPK(copia, cliente);
		HibernateTransaction busca = new HibernateTransaction();
		
		try{
			Session sessao = busca.openSession();
			Reserva reserva = (Reserva) sessao.get(Reserva.class, chave);
			
			if(reserva != null){
				busca.closeSession();
				return reserva;
			}else{
				busca.closeSession();
				throw new ReservaInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ReservaInexistenteException();
		}
	}
	
	@SuppressWarnings("unchecked")
	public Reserva getReserva(long id) throws ReservaInexistenteException {
		HibernateTransaction busca = new HibernateTransaction();
		
		try{
			List<Reserva> lista = (List<Reserva>) busca.openSession().createQuery(String.format("from dados.locacao.Reserva where id=%d", id)).list();
			
			if(!lista.isEmpty()){
				busca.closeSession();
				return lista.get(0);
			}else{
				busca.closeSession();
				throw new ReservaInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ReservaInexistenteException();
		}
	}
	
	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Reserva> getReservas(Copia copia) throws ReservaInexistenteException {
		HibernateTransaction busca = new HibernateTransaction();
		IteratorBidirecional<Reserva> iterador = null;
		
		try{
			List<Reserva> lista = (List<Reserva>) busca.openSession().createQuery(String.format("from dados.locacao.Reserva where id_Copia=%d", copia.getCodigoCopia())).list();
			
			if(!lista.isEmpty()){
				busca.closeSession();
				iterador = new IteratorBidirecional<Reserva>(lista.iterator());
			}else{
				busca.closeSession();
				throw new ReservaInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ReservaInexistenteException();
		}

		return iterador;
	}
	
	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Reserva> getReservas(Cliente cliente) throws ReservaInexistenteException {
		HibernateTransaction busca = new HibernateTransaction();
		IteratorBidirecional<Reserva> iterador = null;
		
		try{
			List<Reserva> lista = (List<Reserva>) busca.openSession().createQuery(String.format("from dados.locacao.Reserva where id_Cliente=%d", cliente.getCpf())).list();
			
			if(!lista.isEmpty()){
				busca.closeSession();
				iterador = new IteratorBidirecional<Reserva>(lista.iterator());
			}else{
				busca.closeSession();
				throw new ReservaInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ReservaInexistenteException();
		}

		return iterador;
	}
	
	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Reserva> getReservaInterceptaPeriodo(Copia copia, Periodo periodo) throws ReservaInexistenteException {
		HibernateTransaction busca = new HibernateTransaction();
		IteratorBidirecional<Reserva> iterador = null;
		
		try{
			List<Reserva> lista = (List<Reserva>) busca.openSession().createQuery(String.format("from dados.locacao.Reserva where id_Copia=%d", copia.getCodigoCopia())).list();
			
			busca.closeSession();
			
			if(!lista.isEmpty()){
				
				List<Reserva> retorno = new ArrayList<Reserva>();
				
				for (Reserva reserva : lista) {
					if(reserva.getPeriodo().intersecta(periodo)){
						retorno.add(reserva);
					}
				}
				
				if(retorno.size() > 0){
					iterador = new IteratorBidirecional<Reserva>(retorno.iterator());
				} else {
					throw new ReservaInexistenteException();
				}
			}else{
				throw new ReservaInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ReservaInexistenteException();
		}

		return iterador;
	}
}
