package armazenamento.filial;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import hibernate.HibernateTransaction;
import negocio.exception.FilialExistenteException;
import negocio.exception.FilialInexistenteException;

import org.hibernate.ObjectNotFoundException;
import org.hibernate.PropertyValueException;
import org.hibernate.Session;
import org.hibernate.exception.ConstraintViolationException;

import util.IteratorBidirecional;

import dados.exception.FilialInvalidaException;
import dados.filial.Filial;

public class RepositorioFilial {

	private static RepositorioFilial instancia;
	private Set<Filial> logadas;
	
	static{
		RepositorioFilial.instancia = new RepositorioFilial();
	}
	
	private RepositorioFilial(){
		this.logadas = new HashSet<Filial>();
	}
	
	public static RepositorioFilial getInstancia(){
		return RepositorioFilial.instancia;
	}
	
	public void cadastrarFilial(Filial filial) throws FilialExistenteException, FilialInvalidaException{
		if(filial != null){
			if(Filial.isValido(filial.getTelefone(), filial.getNome(), filial.getEndereco())){
				HibernateTransaction busca = new HibernateTransaction();
				
				try{
					busca.openSessionBegin();
					Session sessao = busca.getSession();
					
					// AUTO INCREMENT!!!
					try{
						filial.setId((Long)sessao.createQuery("select max(p.id) from dados.filial.Filial p").list().get(0) + 1);
					}catch(NullPointerException e){
						filial.setId(0);
					}
					sessao.save(filial);
					
					busca.closeSessionCommit();
				}catch(ConstraintViolationException e){
					busca.transactionRollBack();
					
					String retorno = e.getSQLException().getMessage();
					
					if(retorno.contains("key 1")){
						throw new FilialExistenteException(filial);
					} 
					
				}catch(PropertyValueException e){
					busca.transactionRollBack();
					throw new FilialInvalidaException();
				}
			} else {
				throw new FilialInvalidaException();
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public void removerFilial(String id) throws FilialInexistenteException{
		if(id != null){
			Filial filial = RepositorioFilial.instancia.getFilial(id);
			
			HibernateTransaction busca = new HibernateTransaction();
			busca.openSessionBegin();
			Session sessao = busca.getSession();
			
			sessao.delete(filial);
			sessao.delete(filial.getTelefone());
			
			busca.closeSessionCommit();
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public boolean existeFilial(String id){
		if(id != null){
			boolean retorno = true;
			
			try{
				RepositorioFilial.instancia.getFilial(id);
			} catch(FilialInexistenteException e){
				retorno = false;
			}
			
			return retorno;
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public Filial getFilial(String id) throws FilialInexistenteException{
		if(id != null){
			HibernateTransaction busca = new HibernateTransaction();
			
			try{
				Session sessao = busca.openSession();
				Filial filial = (Filial) sessao.get(Filial.class, id);
				
				if(filial != null){
					busca.closeSession();
					return filial;
				}else{
					busca.closeSession();
					throw new FilialInexistenteException();
				}
			}catch(ObjectNotFoundException e){
				busca.closeSession();
				throw new FilialInexistenteException();
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	@SuppressWarnings("unchecked")
	public Filial getFilial(long id) throws FilialInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		
		try{
			List<Filial> lista = (List<Filial>) busca.openSession().createQuery(String.format("from dados.filial.Filial where id=%d", id)).list();
			
			if(!lista.isEmpty()){
				busca.closeSession();
				return lista.get(0);
			}else{
				busca.closeSession();
				throw new FilialInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new FilialInexistenteException();
		}
	}
	
	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Filial> getFiliais() throws FilialInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		IteratorBidirecional<Filial> iterador = null;
		
		try{
			List<Filial> lista = (List<Filial>) busca.openSession().createQuery("from dados.filial.Filial").list();
			
			if(!lista.isEmpty()){
				busca.closeSession();
				iterador = new IteratorBidirecional<Filial>(lista.iterator());
			}else{
				busca.closeSession();
				throw new FilialInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new FilialInexistenteException();
		}

		return iterador;
	}

	public void atualizarFilial(Filial atual, Filial dados) throws FilialInvalidaException {
		if(atual != null && dados != null){
			atual.setNome(dados.getNome());
			atual.setTelefone(dados.getTelefone());
			atual.setEndereco(dados.getEndereco());
			
			HibernateTransaction busca = new HibernateTransaction();
			
			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();
				
				sessao.update(atual);
				
				busca.closeSessionCommit();
			}catch(ConstraintViolationException e){
				busca.transactionRollBack();
			}
		} else {
			throw new IllegalArgumentException();
		}
	}
	
	public void logarFilial(Filial filial){
		this.logadas.add(filial);
	}
	
	public void deslogarFilial(String nome){
		Filial f = this.getFilialLogada(nome);
		this.logadas.remove(f);
	}
	
	public boolean estaLogada(String nome){
		return this.getFilialLogada(nome) != null;
	}
	
	private Filial getFilialLogada(String nome){
		Filial retorno = null;
		boolean achou = false;
		
		Iterator<Filial> it = this.logadas.iterator();
		
		while(it.hasNext() && !achou){
			retorno = it.next();
			
			if(retorno.getNome().equals(nome)){
				achou = true;
			}
		}
		
		if(!achou){
			retorno = null;
		}
		
		return retorno;
	}
	
	public IteratorBidirecional<Filial> getFiliaisLogadas(){
		return new IteratorBidirecional<Filial>(this.logadas.iterator());
	}
}
