package armazenamento.produto;

import hibernate.HibernateTransaction;


import java.io.Serializable;
import java.util.Collections;
import java.util.List;

import negocio.exception.AvaliacaoExistenteException;
import negocio.exception.AvaliacaoInexistenteException;

import org.hibernate.ObjectNotFoundException;
import org.hibernate.PropertyValueException;
import org.hibernate.Session;
import org.hibernate.exception.ConstraintViolationException;

import util.IteratorBidirecional;

import dados.exception.AvaliacaoInvalidaException;
import dados.produto.AvaliacaoProduto;
import dados.produto.Produto;
import dados.usuario.Usuario;

public class RepositorioAvaliacaoProduto implements Serializable {

	private static final long serialVersionUID = 1L;

	private static RepositorioAvaliacaoProduto instancia;

	static{
		RepositorioAvaliacaoProduto.instancia = new RepositorioAvaliacaoProduto();
	}

	public RepositorioAvaliacaoProduto() {}

	public static RepositorioAvaliacaoProduto getInstancia(){
		return RepositorioAvaliacaoProduto.instancia;
	}

	public void avaliarProduto(AvaliacaoProduto avaliacao) throws AvaliacaoInvalidaException, AvaliacaoExistenteException{
		HibernateTransaction busca = new HibernateTransaction();

		try{
			busca.openSessionBegin();
			Session sessao = busca.getSession();

			// AUTO INCREMENT!!!
			try{
				avaliacao.setId((Long)sessao.createQuery("select max(p.id) from dados.produto.AvaliacaoProduto p").list().get(0) + 1);
			}catch(NullPointerException e){
				avaliacao.setId(0);
			}

			sessao.save(avaliacao);

			busca.closeSessionCommit();
		}catch(ConstraintViolationException e){
			busca.transactionRollBack();

			String retorno = e.getSQLException().getMessage();

			if(retorno.contains("key 1")){
				throw new AvaliacaoExistenteException(avaliacao.getUsuario().getCpf(), avaliacao.getProduto().getNome());
			}
			
		}catch(PropertyValueException e){
			busca.transactionRollBack();
			throw new AvaliacaoInvalidaException();
		}
	}

	public void removerAvaliacao(Usuario usuario, Produto produto) throws AvaliacaoInexistenteException{
		AvaliacaoProduto avaliacao = this.getAvaliacao(produto, usuario);

		HibernateTransaction busca = new HibernateTransaction();
		busca.openSessionBegin();
		Session sessao = busca.getSession();

		sessao.delete(avaliacao);

		busca.closeSessionCommit();
	}
	
	public void removerAvaliacoes(Usuario usuario) throws AvaliacaoInexistenteException{
		IteratorBidirecional<AvaliacaoProduto> avaliacoes = this.getAvaliacoesUsuario(usuario);
		
		while(avaliacoes.hasNext()){
			this.removerAvaliacao(usuario, avaliacoes.next().getProduto());
		}
	}

	public void removerAvaliacoes(Produto produto) throws AvaliacaoInexistenteException {
		IteratorBidirecional<AvaliacaoProduto> avaliacoes = this.getAvaliacoesProduto(produto);
		
		while(avaliacoes.hasNext()){
			this.removerAvaliacao(avaliacoes.next().getUsuario(), produto);
		}
	}

	public void atualizarAvaliacao(AvaliacaoProduto atual, AvaliacaoProduto dados) throws AvaliacaoInvalidaException{
		atual.setComentarios(dados.getComentarios());
		atual.setNota(dados.getNota());
		atual.setRecomendacao(dados.isRecomendacao());

		HibernateTransaction busca = new HibernateTransaction();

		try{
			busca.openSessionBegin();
			busca.getSession().update(atual);
			busca.closeSessionCommit();
		}catch(Exception e){
			busca.transactionRollBack();
		}
	}

	public boolean existeAvaliacao(Produto produto, Usuario usuario){
		boolean retorno = true;

		try{
			this.getAvaliacao(produto, usuario);
		}catch(AvaliacaoInexistenteException e){
			retorno = false;
		}

		return retorno;
	}

	@SuppressWarnings("unchecked")
	public AvaliacaoProduto getAvaliacao(Produto produto, Usuario usuario) throws AvaliacaoInexistenteException{
		if(produto != null && usuario != null){
			HibernateTransaction busca = new HibernateTransaction();
			AvaliacaoProduto avaliacao = null;

			try{
				Session sessao = busca.openSession();
				List<AvaliacaoProduto> lista = (List<AvaliacaoProduto>) sessao.createQuery(String.format("from dados.produto.AvaliacaoProduto where id_usuario = %d and id_Produto like \'%s\'",usuario.getCpf(), produto.getNome())).list();

				if(lista != null && !lista.isEmpty()){
					avaliacao = lista.get(0);
					busca.closeSession();
					return avaliacao;
				}else{
					busca.closeSession();
					throw new AvaliacaoInexistenteException(usuario.getCpf(), produto.getNome());
				}
			}catch(ObjectNotFoundException e){
				busca.closeSession();
				throw new AvaliacaoInexistenteException(usuario.getCpf(), produto.getNome());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	@SuppressWarnings("unchecked")
	public AvaliacaoProduto getAvaliacaoProduto(long id) throws AvaliacaoInexistenteException {
			HibernateTransaction busca = new HibernateTransaction();

			try{
				Session sessao = busca.openSession();
				List<AvaliacaoProduto> lista = (List<AvaliacaoProduto>) sessao.createQuery(String.format("from dados.produto.AvaliacaoProduto where id = %d",id)).list();

				if(lista != null && !lista.isEmpty()){
					busca.closeSession();
					return lista.get(0);
				}else{
					busca.closeSession();
					throw new AvaliacaoInexistenteException(id);
				}
			}catch(ObjectNotFoundException e){
				busca.closeSession();
				throw new AvaliacaoInexistenteException(id);
			}
	}

	@SuppressWarnings("unchecked")
	public IteratorBidirecional<AvaliacaoProduto> getAvaliacoesProduto(Produto produto) throws AvaliacaoInexistenteException {
		if(produto != null){
			HibernateTransaction busca = new HibernateTransaction();

			try{
				Session sessao = busca.openSession();
				List<AvaliacaoProduto> lista = (List<AvaliacaoProduto>) sessao.createQuery(String.format("from dados.produto.AvaliacaoProduto where id_Produto like \'%s\'",produto.getNome())).list();
				
				Collections.shuffle(lista);

				if(lista != null && !lista.isEmpty()){
					busca.closeSession();
					return new IteratorBidirecional<AvaliacaoProduto>(lista.iterator());
				}else{
					busca.closeSession();
					throw new AvaliacaoInexistenteException(produto.getNome());
				}
			}catch(ObjectNotFoundException e){
				busca.closeSession();
				throw new AvaliacaoInexistenteException(produto.getNome());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	@SuppressWarnings("unchecked")
	public IteratorBidirecional<AvaliacaoProduto> getAvaliacoesUsuario(Usuario usuario) throws AvaliacaoInexistenteException {
		if(usuario != null){
			HibernateTransaction busca = new HibernateTransaction();

			try{
				Session sessao = busca.openSession();
				List<AvaliacaoProduto> lista = (List<AvaliacaoProduto>) sessao.createQuery(String.format("from dados.produto.AvaliacaoProduto where id_usuario = %d",usuario.getCpf())).list();

				if(lista != null && !lista.isEmpty()){
					busca.closeSession();
					return new IteratorBidirecional<AvaliacaoProduto>(lista.iterator());
				}else{
					busca.closeSession();
					throw new AvaliacaoInexistenteException(usuario.getCpf());
				}
			}catch(ObjectNotFoundException e){
				busca.closeSession();
				throw new AvaliacaoInexistenteException(usuario.getCpf());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

}
