package armazenamento.produto;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.StringTokenizer;

import negocio.exception.ClassificacaoInexistenteException;
import negocio.exception.CopiaExistenteException;
import negocio.exception.CopiaInexistenteException;
import negocio.exception.LinkFotoExistenteException;
import negocio.exception.ProdutoExistenteException;
import negocio.exception.ProdutoInexistenteException;

import org.hibernate.ObjectNotFoundException;
import org.hibernate.PropertyValueException;
import org.hibernate.Session;
import org.hibernate.exception.ConstraintViolationException;

import util.ComparadorProdutoNome;
import util.ComparadorProdutoNota;
import util.IteratorBidirecional;

import dados.exception.FilialInvalidaException;
import dados.exception.PrecoInvalidoException;
import dados.exception.ProdutoInvalidoException;
import dados.filial.Filial;
import dados.locacao.Copia;
import dados.produto.ClassificacaoPreco;
import dados.produto.CriterioBusca;
import dados.produto.LinkFoto;
import dados.produto.Produto;
import dados.produto.TipoProduto;

import hibernate.HibernateTransaction;

public class RepositorioProduto {

	private static RepositorioProduto instancia;

	static{
		RepositorioProduto.instancia = new RepositorioProduto();
	}

	public RepositorioProduto() {}

	public static RepositorioProduto getInstancia(){
		return RepositorioProduto.instancia;
	}

	public boolean existeClassificacao(String descricao){
		if(descricao != null){
			boolean existe = true;

			try{
				getClassificacao(descricao);
			}catch(ClassificacaoInexistenteException e){
				existe = false;
			}

			return existe;
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void inserirClassificacao(ClassificacaoPreco c){
		if(c != null){
			HibernateTransaction busca = new HibernateTransaction();

			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();

				// AUTO INCREMENT!!!
				try{
					c.setId((Long)sessao.createQuery("select max(p.id) from dados.produto.ClassificacaoPreco p").list().get(0) + 1);
				}catch(NullPointerException e){
					c.setId(0);
				}

				sessao.save(c);

				busca.closeSessionCommit();
			}catch(ConstraintViolationException e){
				busca.transactionRollBack();
			}catch(PropertyValueException e){
				busca.transactionRollBack();
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public ClassificacaoPreco getClassificacao(String descricao) throws ClassificacaoInexistenteException{
		if(descricao != null){
			ClassificacaoPreco classificacao;

			HibernateTransaction busca = new HibernateTransaction();

			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();

				classificacao = (ClassificacaoPreco) sessao.get(ClassificacaoPreco.class, descricao);

				busca.closeSessionCommit();

				if(classificacao != null){
					return classificacao;
				}else{
					throw new ClassificacaoInexistenteException(descricao);
				}
			}catch(ObjectNotFoundException e){
				busca.transactionRollBack();
				throw new ClassificacaoInexistenteException(descricao);
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void removerClassificacao(String descricao) throws ClassificacaoInexistenteException {
		if(descricao != null){
			ClassificacaoPreco classificacao = this.getClassificacao(descricao);

			HibernateTransaction busca = new HibernateTransaction();

			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();

				sessao.delete(classificacao);

				busca.closeSessionCommit();
			}catch(ObjectNotFoundException e){
				busca.transactionRollBack();
				throw new ClassificacaoInexistenteException(descricao);
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void atualizarClassificacao(ClassificacaoPreco classificacao) throws PrecoInvalidoException, ClassificacaoInexistenteException {
		if(classificacao != null){
			ClassificacaoPreco c = this.getClassificacao(classificacao.getDescricao());

			c.setPreco(classificacao.getPreco());

			HibernateTransaction busca = new HibernateTransaction();

			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();

				sessao.update(classificacao);

				busca.closeSessionCommit();
			}catch(ObjectNotFoundException e){
				busca.transactionRollBack();
				throw new ClassificacaoInexistenteException(classificacao.getDescricao());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void inserirProduto(Produto produto) throws ProdutoExistenteException, ProdutoInvalidoException{
		if(produto != null){
			HibernateTransaction busca = new HibernateTransaction();

			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();

				// AUTO INCREMENT!!!
				try{
					produto.setId((Long)sessao.createQuery("select max(p.id) from dados.produto.Produto p").list().get(0) + 1);
				}catch(NullPointerException e){
					produto.setId(0);
				}

				sessao.save(produto);

				busca.closeSessionCommit();
			}catch(ConstraintViolationException e){
				busca.transactionRollBack();
				throw new ProdutoExistenteException(produto.getNome());
			}catch(PropertyValueException e){
				busca.transactionRollBack();
				throw new ProdutoInvalidoException(produto);
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void removerProduto(String nome) throws ProdutoInexistenteException{
		Produto produto = this.getProduto(nome);

		HibernateTransaction busca = new HibernateTransaction();
		busca.openSessionBegin();
		Session sessao = busca.getSession();

		sessao.delete(produto);

		busca.closeSessionCommit();
	}

	public void alterarDados(Produto atual, Produto dados) throws ProdutoInvalidoException{
		if(dados != null){
			HibernateTransaction busca = new HibernateTransaction();
			atual.setDados(dados);
			
			try{
				busca.openSessionBegin();
				Session sessao = busca.getSession();

				sessao.update(atual);

				busca.closeSessionCommit();
			}catch(PropertyValueException e){
				throw new ProdutoInvalidoException(dados);
			}
		}else{
			throw new IllegalArgumentException();
		}
	}


	public boolean existeProduto(String nome){
		if(nome != null){
			boolean retorno = true;

			try{
				this.getProduto(nome);
			}catch(ProdutoInexistenteException e){
				retorno = false;
			}

			return retorno;
		}else{
			throw new IllegalArgumentException();
		}
	}

	@SuppressWarnings("unchecked")
	public Produto getProduto(String nome) throws ProdutoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();

		try{
			List<Produto> lista = (List<Produto>) busca.openSession().createQuery(String.format("from dados.produto.Produto where nome like '%s'",nome)).list();

			if(!lista.isEmpty()){
				busca.closeSession();
				return lista.get(0);
			}else{
				busca.closeSession();
				throw new ProdutoInexistenteException(nome);
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ProdutoInexistenteException(nome);
		}
	}
	
	@SuppressWarnings("unchecked")
	public Produto getProduto(long id) throws ProdutoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();

		try{
			List<Produto> lista = (List<Produto>) busca.openSession().createQuery(String.format("from dados.produto.Produto where id=%d",id)).list();

			if(!lista.isEmpty()){
				busca.closeSession();
				return lista.get(0);
			}else{
				busca.closeSession();
				throw new ProdutoInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ProdutoInexistenteException();
		}
	}

	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Produto> getProdutosTipo(TipoProduto tipo) throws ProdutoInexistenteException{
		IteratorBidirecional<Produto> iterator = null;

		if(tipo != null){
			HibernateTransaction busca = new HibernateTransaction();

			try{
				List<Produto> lista = (List<Produto>) busca.openSession().createQuery(String.format("from dados.produto.%s", tipo.getNome())).list();;

				if(!lista.isEmpty()){
					busca.closeSession();
					iterator = new IteratorBidirecional<Produto>(lista.iterator());
				}else{
					busca.closeSession();
					throw new ProdutoInexistenteException();
				}
			}catch(ObjectNotFoundException e){
				busca.closeSession();
				throw new ProdutoInexistenteException();
			}
		} else {
			throw new IllegalArgumentException();
		}

		return iterator;
	}

	private String printWhereOrAnd(String query, boolean where){
		if(where){
			query = query.concat(" where ");
		}else{
			query = query.concat(" and ");
		}

		return query;
	}

	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Produto> getProdutosComRestricoes(CriterioBusca restricoes) throws ProdutoInexistenteException{
		if(restricoes != null){
			HibernateTransaction busca = new HibernateTransaction();
			IteratorBidirecional<Produto> iterador = null;

			try{
				String query = "from dados.produto.";
				boolean where = true;

				if(restricoes.getTipo() != null){
					query = query.concat(restricoes.getTipo().getNome());
				}else{
					query = query.concat("Produto");
				}

				if(restricoes.getAno() != null){
					query = this.printWhereOrAnd(query, where);
					query = query.concat(String.format("AnoLancamento = %d", restricoes.getAno()));
					where = false;
				}

				if(restricoes.getCensura() != null){
					query = this.printWhereOrAnd(query, where);
					query = query.concat(String.format("Censura like '%s'", restricoes.getCensura().toLowerCase()));
					where = false;
				}

				if(restricoes.getClassificacaoPreco() != null){
					query = this.printWhereOrAnd(query, where);
					query = query.concat(String.format("id_Classificacao like '%s'", restricoes.getClassificacaoPreco()));				
					where = false;
				}

				if(restricoes.getEmpresa() != null){
					query = this.printWhereOrAnd(query, where);
					query = query.concat(String.format("EmpresaProdutora like '%s'", restricoes.getEmpresa().toLowerCase()));
					where = false;
				}

				if(restricoes.getGenero() != null){
					query = this.printWhereOrAnd(query, where);
					query = query.concat(String.format("Genero like '%s'", restricoes.getGenero().toLowerCase()));
					where = false;
				}

				List<Produto> lista = (List<Produto>) busca.openSession().createQuery(query).list();

				busca.closeSession();

				if(!lista.isEmpty()){
					if(restricoes.getNome() != null){
						ArrayList<String> padroes = new ArrayList<String>();
						ArrayList<String> nomes = new ArrayList<String>();
						StringTokenizer tool = new StringTokenizer(restricoes.getNome().toLowerCase());
						StringTokenizer tool2;

						while(tool.hasMoreTokens()){
							padroes.add(tool.nextToken());
						}

						boolean contem;

						for(int i = 0; i < lista.size(); i++){
							contem = false;
							tool2 = new StringTokenizer(lista.get(i).getNome().toLowerCase());
							nomes.clear();
							
							while(tool2.hasMoreTokens()){
								nomes.add(tool2.nextToken());
							}

							for(int k = nomes.size()-1; k >= 0; k--){
								for(int j = padroes.size()-1; j >= 0; j--){
									if(nomes.get(k).equalsIgnoreCase(padroes.get(j))){
										contem = true;
									}
								}
							}
							
							if(!contem){
								lista.remove(i);
								i--;
							}
						}

						Collections.sort(lista, new ComparadorProdutoNome(padroes));
					}

					if(!lista.isEmpty()){
						iterador = new IteratorBidirecional<Produto>(lista.iterator());
					} else{
						throw new ProdutoInexistenteException();
					}
				}else{
					throw new ProdutoInexistenteException();
				}

				return iterador;
			}catch(ObjectNotFoundException e){
				busca.closeSession();
				throw new ProdutoInexistenteException();
			}
		} else throw new IllegalArgumentException();
	}

	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Produto> getProdutos() throws ProdutoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		IteratorBidirecional<Produto> iterator = null;

		try{
			List<Produto> lista = (List<Produto>) busca.openSession().createQuery("from dados.produto.Produto").list();

			if(!lista.isEmpty()){
				busca.closeSession();
				iterator = new IteratorBidirecional<Produto>(lista.iterator());
			}else{
				busca.closeSession();
				throw new ProdutoInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ProdutoInexistenteException();
		}

		return iterator;
	}

	public void inserirLinkFoto(String nome, LinkFoto foto) throws LinkFotoExistenteException, ProdutoInexistenteException{
		Produto p = this.getProduto(nome);

		HibernateTransaction busca = new HibernateTransaction();
		Produto atual = this.getProduto(nome);
		
		p.inserirLinkFoto(foto);

		try{
			this.alterarDados(atual, p);
		} catch (ProdutoInvalidoException e) {
			//NAO deve acontecer
		} catch(Exception e){
			busca.transactionRollBack();
		}
	}

	public void inserirCopia(Copia copia) throws CopiaExistenteException, ProdutoInexistenteException{
		Produto produto = this.getProduto(copia.getProduto().getNome());
		produto.inserirCopia(copia);

		HibernateTransaction busca = new HibernateTransaction();

		try{
			busca.openSessionBegin();
			busca.getSession().update(produto);
			busca.closeSessionCommit();
		}catch(Exception e){
			busca.transactionRollBack();
		}
	}

	public void removerCopia(Copia copia) throws CopiaInexistenteException, ProdutoInexistenteException{
		Produto produto = this.getProduto(copia.getProduto().getNome());
		produto.removerCopia(copia.getCodigoCopia());

		HibernateTransaction busca = new HibernateTransaction();

		try{
			busca.openSessionBegin();
			busca.getSession().delete(copia);
			busca.getSession().update(produto);
			busca.closeSessionCommit();
		}catch(Exception e){
			busca.transactionRollBack();
		}
	}

	public void mudarFilialCopia(Produto produto, int codigoCopia, Filial filial) throws CopiaInexistenteException, FilialInvalidaException{
		if(produto != null && filial != null){
			produto.mudarFilialCopia(codigoCopia, filial);
			Copia copia = produto.getCopia(codigoCopia);

			HibernateTransaction busca = new HibernateTransaction();

			try{
				busca.openSessionBegin();
				busca.getSession().update(copia);
				busca.getSession().update(produto);
				busca.closeSessionCommit();
			}catch(Exception e){
				busca.transactionRollBack();
			}
		} else {
			throw new IllegalArgumentException();
		}
	}

	@SuppressWarnings("unchecked")
	public IteratorBidirecional<Produto> getProdutosRecomendados() throws ProdutoInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		List<Produto> lista = null;

		try{
			Session sessao = busca.openSession();
			lista = (List<Produto>) sessao.createQuery("from dados.produto.Produto").list();

			Collections.sort(lista, new ComparadorProdutoNota());

			busca.closeSession();
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ProdutoInexistenteException();
		}

		if(!lista.isEmpty()){
			return new IteratorBidirecional<Produto>(lista.iterator());
		} else {
			throw new ProdutoInexistenteException();
		}
	}

	@SuppressWarnings("unchecked")
	public IteratorBidirecional<ClassificacaoPreco> getClassificacoes() throws ClassificacaoInexistenteException {
		HibernateTransaction busca = new HibernateTransaction();
		List<ClassificacaoPreco> lista = null;

		try{
			Session sessao = busca.openSession();
			lista = (List<ClassificacaoPreco>) sessao.createQuery("from dados.produto.ClassificacaoPreco").list();

			busca.closeSession();
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new ClassificacaoInexistenteException();
		}

		return new IteratorBidirecional<ClassificacaoPreco>(lista.iterator());
	}
	
	@SuppressWarnings("unchecked")
	public Copia buscarCopia(int codigoCopia) throws CopiaInexistenteException{
		HibernateTransaction busca = new HibernateTransaction();
		List<Copia> lista = null;

		try{
			Session sessao = busca.openSession();
			lista = (List<Copia>) sessao.createQuery(String.format("from dados.locacao.Copia where id_Copia = %d", codigoCopia)).list();

			busca.closeSession();
			
			if(!lista.isEmpty()){
				return lista.get(0);
			} else {
				throw new CopiaInexistenteException();
			}
		}catch(ObjectNotFoundException e){
			busca.closeSession();
			throw new CopiaInexistenteException();
		}
	}

	public boolean existeCopia(int codigoCopia) {
		try {
			this.buscarCopia(codigoCopia);
			return true;
		} catch (CopiaInexistenteException e) {
			return false;
		}
	}

	@SuppressWarnings("unchecked")
	public ClassificacaoPreco getClassificacao(long id) throws ClassificacaoInexistenteException {
		HibernateTransaction busca = new HibernateTransaction();

		try{
			busca.openSessionBegin();
			Session sessao = busca.getSession();

			List<ClassificacaoPreco> lista = (List<ClassificacaoPreco>) sessao.createQuery(String.format("from dados.produto.ClassificacaoPreco where id=%d", id)).list();

			busca.closeSessionCommit();

			if(!lista.isEmpty()){
				return lista.get(0);
			}else{
				throw new ClassificacaoInexistenteException("");
			}
		}catch(ObjectNotFoundException e){
			busca.transactionRollBack();
			throw new ClassificacaoInexistenteException("");
		}
	}

}
