package negocio;

import util.IteratorBidirecional;

import negocio.exception.SenhaInvalidaException;
import negocio.exception.UsuarioExistenteException;
import negocio.exception.UsuarioInexistenteException;
import negocio.exception.UsuarioLogadoException;
import negocio.exception.UsuarioNaoLogadoException;
import negocio.exception.UsuarioSemPermissaoException;
import armazenamento.usuario.RepositorioUsuario;
import dados.usuario.Cliente;
import dados.usuario.Telefone;
import dados.usuario.Usuario;

public class GerenciadorUsuario {

	private static GerenciadorUsuario instancia;
	private RepositorioUsuario repositorio;

	static{
		GerenciadorUsuario.instancia = new GerenciadorUsuario();
	}

	private GerenciadorUsuario(){
		this.repositorio = RepositorioUsuario.getInstancia();
	}

	public static GerenciadorUsuario getInstancia(){
		return GerenciadorUsuario.instancia;
	}

	public long logar(long cpf, String senha) throws UsuarioLogadoException, UsuarioInexistenteException, SenhaInvalidaException{
		if(this.repositorio.existeUsuario(cpf)){
			Usuario usuario = this.repositorio.getUsuario(cpf);

			if(!this.repositorio.estaLogado(usuario)){
				if(usuario.getSenha().equals(senha)){
					return this.repositorio.logarUsuario(usuario);
				} else {
					throw new SenhaInvalidaException(cpf, senha);
				}
			}else{
				throw new UsuarioLogadoException(usuario);
			}
		}else{
			throw new UsuarioInexistenteException(cpf);
		}
	}

	public void deslogar(long sessao) throws UsuarioNaoLogadoException{
		if(this.repositorio.estaLogado(sessao)){
			this.repositorio.deslogarSessao(sessao);
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	public boolean estaLogado(Usuario user){
		return this.repositorio.estaLogado(user);
	}

	public boolean estaLogado(long sessao){
		return this.repositorio.estaLogado(sessao);
	}

	public Usuario getUsuarioLogado(long sessao) throws UsuarioNaoLogadoException{
		if(this.repositorio.estaLogado(sessao)){
			return this.repositorio.getUsuarioLogado(sessao);
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	public void cadastrarCliente(Cliente user, Telefone[] telefone) throws UsuarioExistenteException{
		if(user != null){
			if(!this.repositorio.existeUsuario(user.getCpf())){
				this.repositorio.cadastrarUsuario(user);
				
				try {
					this.alterarTelefones(user.getCpf(), telefone);
				} catch (UsuarioInexistenteException e) {
					//NAO deve ocorrer
				}
			}else{
				throw new UsuarioExistenteException(user.getCpf());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	private void alterarTelefones(Long cpf, Telefone[] telefone) throws UsuarioInexistenteException{
		if(this.repositorio.existeUsuario(cpf)){
			Cliente user = (Cliente) this.buscarUsuario(cpf);
			user.setTelefones(telefone);
			
			try {
				this.alterarUsuario(user);
			} catch (UsuarioExistenteException e) {
				//NAO debve ocorrer
			}
		} else {
			throw new UsuarioInexistenteException();
		}
	}

	/**
	 * Realiza o cadastro de um usurio online onde  necessario
	 * ter permisso para realizar a operao.
	 * @param user
	 * @param sessao
	 * @throws UsuarioSemPermissaoException
	 * @throws UsuarioExistenteException
	 * @throws UsuarioInvalidoException
	 * @throws UsuarioNaoLogadoException 
	 */
	public void cadastrarUsuario(Usuario user, long sessao) throws UsuarioSemPermissaoException, UsuarioExistenteException, UsuarioNaoLogadoException{
		if(user != null){
			if(!this.repositorio.existeUsuario(user.getCpf())){

				if(this.repositorio.estaLogado(sessao)){
					Usuario usuario = this.repositorio.getUsuarioLogado(sessao);

					if(usuario.temPermissaoCadastro(user)){
						this.repositorio.cadastrarUsuario(user);
					}else{
						throw new UsuarioSemPermissaoException(usuario);
					}
				}else{
					throw new UsuarioNaoLogadoException();
				}
			}else{
				throw new UsuarioExistenteException(user.getCpf());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void descadastrarUsuario(Usuario user, long sessao, String senha) throws UsuarioInexistenteException, UsuarioSemPermissaoException, UsuarioNaoLogadoException, SenhaInvalidaException {
		if(user != null && senha != null){
			if(this.repositorio.existeUsuario(user.getCpf())){

				if(this.repositorio.estaLogado(sessao)){
					Usuario usuario = this.repositorio.getUsuarioLogado(sessao);

					if(usuario.temPermissaoRemocao(user)){
						if(usuario.validarSenha(senha)){
							this.repositorio.removerUsuario(user.getCpf());
						}else{
							throw new SenhaInvalidaException(usuario.getCpf(), senha);
						}
					}else{
						throw new UsuarioSemPermissaoException(usuario);
					}
				}else{
					throw new UsuarioNaoLogadoException();
				}
			}else{
				throw new UsuarioInexistenteException(user.getCpf());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public Usuario buscarUsuario(long cpf) throws UsuarioInexistenteException{
		if(this.repositorio.existeUsuario(cpf)){
			return this.repositorio.getUsuario(cpf);
		}else{
			throw new UsuarioInexistenteException(cpf); 
		}
	}

	public void alterarUsuario(long sessao, Usuario dados) throws UsuarioInexistenteException, UsuarioSemPermissaoException, UsuarioNaoLogadoException, UsuarioExistenteException{
		if(dados != null){
			if(this.repositorio.estaLogado(sessao)){
				Usuario transacao = this.repositorio.getUsuarioLogado(sessao);
				Usuario user = this.buscarUsuario(dados.getCpf());

				if(transacao.temPermissaoAlteracao(user)){
					this.alterarUsuario(dados);
				}else{
					throw new UsuarioSemPermissaoException(transacao);
				}
			}else{
				throw new UsuarioNaoLogadoException();
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void alterarUsuario(Usuario dados) throws UsuarioInexistenteException, UsuarioExistenteException{
		if(dados != null){
			this.repositorio.alterarDados(this.repositorio.getUsuario(dados.getCpf()), dados);
		}else{
			throw new IllegalArgumentException();
		}
	}

	public IteratorBidirecional<Usuario> getUsuarios() throws UsuarioInexistenteException{
		return this.repositorio.getUsuarios();
	}

	public void descadastrarUsuario(Usuario user) throws UsuarioInexistenteException {
		if(this.repositorio.existeUsuario(user.getCpf())){
			this.repositorio.removerUsuario(user.getCpf());
		}else{
			throw new UsuarioInexistenteException(user.getCpf());
		}

	}

	public void alterarTelefonesCliente(long sessao, long cpf, Telefone[] telefones) throws UsuarioNaoLogadoException, UsuarioInexistenteException{
		if(this.estaLogado(sessao)){
			this.alterarTelefones(cpf, telefones);
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}
}
