package negocio.produto;

import util.IteratorBidirecional;

import negocio.exception.ClassificacaoExistenteException;
import negocio.exception.ClassificacaoInexistenteException;
import negocio.exception.CopiaExistenteException;
import negocio.exception.CopiaInexistenteException;
import negocio.exception.LinkFotoExistenteException;
import negocio.exception.ProdutoExistenteException;
import negocio.exception.ProdutoInexistenteException;
import dados.exception.FilialInvalidaException;
import dados.exception.PrecoInvalidoException;
import dados.exception.ProdutoInvalidoException;
import dados.filial.Filial;
import dados.locacao.Copia;
import dados.produto.ClassificacaoPreco;
import dados.produto.CriterioBusca;
import dados.produto.LinkFoto;
import dados.produto.Produto;
import dados.produto.TipoProduto;
import armazenamento.produto.RepositorioProduto;

public class GerenciadorProduto {

	private static GerenciadorProduto instancia;
	private RepositorioProduto repositorio;

	static{
		GerenciadorProduto.instancia = new GerenciadorProduto();
	}

	private GerenciadorProduto(){
		this.repositorio = RepositorioProduto.getInstancia();
	}

	public static GerenciadorProduto getInstancia(){
		return GerenciadorProduto.instancia;
	}

	public void inserirProduto(Produto produto) throws ProdutoExistenteException, ProdutoInvalidoException{
		if(produto != null){
			if(!this.repositorio.existeProduto(produto.getNome())){
				this.repositorio.inserirProduto(produto);
			} else{
				throw new ProdutoExistenteException(produto.getNome());
			}
		} else {
			throw new IllegalArgumentException();
		}
	}

	public void removerProduto(String nome) throws ProdutoInexistenteException{
		if(this.repositorio.existeProduto(nome)){
			this.repositorio.removerProduto(nome);
		} else {
			throw new ProdutoInexistenteException();
		}
	}

	public boolean existeProduto(String nome){
		if(nome != null){
			return this.repositorio.existeProduto(nome);
		}else{
			throw new IllegalArgumentException();
		}
	}

	public Produto buscarProduto(String nome) throws ProdutoInexistenteException{
		if(nome != null){
			if(this.repositorio.existeProduto(nome)){
				return this.repositorio.getProduto(nome);
			} else{
				throw new ProdutoInexistenteException();
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public Produto buscarProduto(long id) throws ProdutoInexistenteException{
		return this.repositorio.getProduto(id);
	}
	
	public ClassificacaoPreco getClassificacaoPreco(String descricao) throws ClassificacaoInexistenteException{
		if(descricao != null){
			if(this.repositorio.existeClassificacao(descricao)){
				return this.repositorio.getClassificacao(descricao);
			}else{
				throw new ClassificacaoInexistenteException(descricao);
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public void inserirClassificacao(ClassificacaoPreco classificacao) throws ClassificacaoExistenteException{
		if(classificacao != null){
			if(!this.repositorio.existeClassificacao(classificacao.getDescricao())){
				this.repositorio.inserirClassificacao(classificacao);
			}else{
				throw new ClassificacaoExistenteException(classificacao.getDescricao());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}

	public void removerClassificacao(String descricao) throws ClassificacaoInexistenteException {
		if(descricao != null){
			if(this.repositorio.existeClassificacao(descricao)){
				this.repositorio.removerClassificacao(descricao);
			}else{
				throw new ClassificacaoInexistenteException(descricao);
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public void atualizarClassificacao(ClassificacaoPreco classificacao) throws PrecoInvalidoException, ClassificacaoInexistenteException {
		if(classificacao != null){
			if(this.repositorio.existeClassificacao(classificacao.getDescricao())){
				this.repositorio.atualizarClassificacao(classificacao);
			}else{
				throw new ClassificacaoInexistenteException(classificacao.getDescricao());
			}
		}else{
			throw new IllegalArgumentException();
		}
	}
	
	public IteratorBidirecional<Produto> buscarProdutoTipo(TipoProduto tipo) throws ProdutoInexistenteException{
		if(tipo != null){
			return this.repositorio.getProdutosTipo(tipo);
		} else {
			throw new IllegalArgumentException();
		}
	}

	public IteratorBidirecional<Produto> buscarProdutoRestricoes(CriterioBusca restricoes) throws ProdutoInexistenteException{
		if(restricoes != null){
			return this.repositorio.getProdutosComRestricoes(restricoes);
		} else {
			throw new IllegalArgumentException();
		}
	}

	public void alterarDadosProduto(Produto dados) throws ProdutoInexistenteException, ProdutoInvalidoException, ProdutoExistenteException{
		if(dados != null && dados != null){
			if(this.repositorio.existeProduto(dados.getNome())){
				Produto atual = this.repositorio.getProduto(dados.getNome());
				if(dados.getNota() == 0){
					dados.setNota(atual.getNota());
				}
				
				this.repositorio.alterarDados(atual, dados);
			} else {
				throw new ProdutoInexistenteException();
			}
		} else {
			throw new IllegalArgumentException();
		}
	}

	public void inserirCopia(Copia copia) throws CopiaExistenteException, ProdutoInexistenteException{
		if(copia != null){
			if(this.repositorio.existeProduto(copia.getProduto().getNome())){
				if(!this.repositorio.existeCopia(copia.getCodigoCopia())){
					this.repositorio.inserirCopia(copia);
				} else {
					throw new CopiaExistenteException(copia);
				}
			} else {
				throw new ProdutoInexistenteException();
			}
		} else {
			throw new IllegalArgumentException();
		}
	}

	public void removerCopia(Copia copia) throws ProdutoInexistenteException, CopiaInexistenteException{
		if(copia != null){
			if(this.repositorio.existeProduto(copia.getProduto().getNome())){
				this.repositorio.removerCopia(copia);
			} else {
				throw new ProdutoInexistenteException();
			}
		} else {
			throw new IllegalArgumentException();
		}
	}
	
	public void inserirLinkFoto(String nome, LinkFoto link) throws LinkFotoExistenteException, ProdutoInexistenteException{
		this.repositorio.inserirLinkFoto(nome, link);
	}

	public void mudarFilialDeCopia(Copia copia, Filial filial) throws ProdutoInexistenteException, CopiaInexistenteException, FilialInvalidaException{
		if(copia != null && filial != null){
			if(this.repositorio.existeProduto(copia.getProduto().getNome())){
				Produto produto = this.repositorio.getProduto(copia.getProduto().getNome());
				this.repositorio.mudarFilialCopia(produto, copia.getCodigoCopia(), filial);
			} else {
				throw new ProdutoInexistenteException();
			}
		} else {
			throw new IllegalArgumentException();
		}
	}

	public IteratorBidirecional<Produto> getProdutos() throws ProdutoInexistenteException{
		return this.repositorio.getProdutos();
	}
	
	public IteratorBidirecional<Produto> getProdutosRecomendados() throws ProdutoInexistenteException{
		return this.repositorio.getProdutosRecomendados();
	}

	public IteratorBidirecional<ClassificacaoPreco> getClassificacoes() throws ClassificacaoInexistenteException {
		return this.repositorio.getClassificacoes();
	}
	
	public Copia buscarCopia(int codigoCopia) throws CopiaInexistenteException{
		return this.repositorio.buscarCopia(codigoCopia);
	}

	public ClassificacaoPreco getClassificacaoPreco(long id) throws ClassificacaoInexistenteException {
		return this.repositorio.getClassificacao(id);
	}
}
