package servicos;

import java.net.MalformedURLException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RMISecurityManager;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Timer;

import util.IteratorBidirecional;

import negocio.GerenciadorFilial;
import negocio.GerenciadorPlano;
import negocio.GerenciadorUsuario;
import negocio.exception.AvaliacaoInexistenteException;
import negocio.exception.CopiaInexistenteException;
import negocio.exception.CopiaLocadaNoPeriodoException;
import negocio.exception.CopiaReservadaNoPeriodoException;
import negocio.exception.FilialInexistenteException;
import negocio.exception.FilialNaoLogadaException;
import negocio.exception.LocacaoExistenteException;
import negocio.exception.LocacaoInexistenteException;
import negocio.exception.PlanoInexistenteException;
import negocio.exception.ProdutoInexistenteException;
import negocio.exception.PromocaoInexistenteException;
import negocio.exception.ReservaExistenteException;
import negocio.exception.ReservaInexistenteException;
import negocio.exception.SenhaInvalidaException;
import negocio.exception.UsuarioExistenteException;
import negocio.exception.UsuarioInexistenteException;
import negocio.exception.UsuarioLogadoException;
import negocio.exception.UsuarioNaoLogadoException;
import negocio.exception.UsuarioSemPermissaoException;
import negocio.locacao.GerenciadorLocacao;
import negocio.locacao.GerenciadorPromocao;
import negocio.locacao.GerenciadorReserva;
import negocio.produto.GerenciadorAvaliacaoProduto;
import negocio.produto.GerenciadorProduto;

import dados.exception.FilialInvalidaException;
import dados.exception.LocacaoInvalidaException;
import dados.filial.Filial;
import dados.locacao.Copia;
import dados.locacao.Locacao;
import dados.locacao.Periodo;
import dados.locacao.Promocao;
import dados.locacao.Reserva;
import dados.produto.CriterioBusca;
import dados.produto.Produto;
import dados.produto.TipoProduto;
import dados.usuario.Administrador;
import dados.usuario.Cliente;
import dados.usuario.Dependente;
import dados.usuario.Funcionario;
import dados.usuario.Plano;
import dados.usuario.Telefone;
import dados.usuario.Usuario;

public class InterfaceGeralImp extends UnicastRemoteObject implements InterfaceGeral {

	private static final long serialVersionUID = 1L;

	protected GerenciadorLocacao gerenciadorLocacao;
	protected GerenciadorPlano gerenciadorPlano;
	protected GerenciadorProduto gerenciadorProduto;
	protected GerenciadorPromocao gerenciadorPromocao;
	protected GerenciadorReserva gerenciadorReserva;
	protected GerenciadorUsuario gerenciadorUsuario;
	protected GerenciadorFilial gerenciadorFilial;
	protected GerenciadorAvaliacaoProduto gerenciadorAvaliacaoProduto;

	public InterfaceGeralImp() throws RemoteException {
		super();

		this.gerenciadorLocacao = GerenciadorLocacao.getInstancia();
		this.gerenciadorPlano = GerenciadorPlano.getInstancia();
		this.gerenciadorProduto = GerenciadorProduto.getInstancia();
		this.gerenciadorPromocao = GerenciadorPromocao.getInstancia();
		this.gerenciadorReserva = GerenciadorReserva.getInstancia();
		this.gerenciadorUsuario = GerenciadorUsuario.getInstancia();
		this.gerenciadorFilial = GerenciadorFilial.getInstancia();
		this.gerenciadorAvaliacaoProduto = GerenciadorAvaliacaoProduto.getInstancia();
	}

	@Override
	public synchronized void alterarDadosUsuario(long sessao, Usuario dados) throws UsuarioNaoLogadoException, UsuarioInexistenteException, UsuarioExistenteException, UsuarioSemPermissaoException, RemoteException {
		this.gerenciadorUsuario.alterarUsuario(sessao, dados);
	}
	
	@Override
	public synchronized Locacao buscarLocacao(long id) throws LocacaoInexistenteException, RemoteException{
		return this.gerenciadorLocacao.buscarLocacao(id);
	}

	@Override
	public synchronized Locacao buscarLocacao(long sessao, Copia copia) throws RemoteException, RemoteException, UsuarioNaoLogadoException, UsuarioSemPermissaoException, LocacaoInexistenteException {
		if(this.gerenciadorUsuario.estaLogado(sessao)){
			Usuario transicao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
			Locacao locacao = this.gerenciadorLocacao.buscarLocacao(copia);

			if(transicao instanceof Cliente){
				if(locacao.getCliente().equals(transicao)){
					return locacao;
				} else {
					throw new UsuarioSemPermissaoException(transicao);
				}
			} else if(transicao instanceof Dependente){
				if(locacao.getCliente().equals(((Dependente) transicao).getCliente())){
					return locacao;
				} else {
					throw new UsuarioSemPermissaoException(transicao);
				}
			} else {
				return locacao;
			}
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	@Override
	public synchronized IteratorBidirecional<Locacao> buscarLocacoes(long sessao, Cliente cliente) throws RemoteException, LocacaoInexistenteException, UsuarioSemPermissaoException, UsuarioNaoLogadoException {
		if(this.gerenciadorUsuario.estaLogado(sessao)){
			Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
			if(transacao instanceof Administrador || transacao instanceof Funcionario){
				return this.gerenciadorLocacao.buscarLocacoes(cliente);
			} else {
				throw new UsuarioSemPermissaoException(transacao);
			}
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	@Override
	public synchronized Plano buscarPlano(long id) throws PlanoInexistenteException, RemoteException{
		return this.gerenciadorPlano.buscarPlano(id);
	}
	
	@Override
	public synchronized Plano buscarPlano(String nome) throws RemoteException, PlanoInexistenteException {
		return this.gerenciadorPlano.buscarPlano(nome);
	}
	
	@Override
	public synchronized Produto buscarProduto(long id) throws ProdutoInexistenteException, RemoteException{
		return this.gerenciadorProduto.buscarProduto(id);
	}

	@Override
	public synchronized Produto buscarProduto(String nomeProduto) throws RemoteException, ProdutoInexistenteException {
		return this.gerenciadorProduto.buscarProduto(nomeProduto);
	}

	@Override
	public synchronized IteratorBidirecional<Produto> buscarProdutoRestricoes(CriterioBusca restricoes) throws RemoteException, ProdutoInexistenteException {
		return this.gerenciadorProduto.buscarProdutoRestricoes(restricoes);
	}

	@Override
	public synchronized IteratorBidirecional<Produto> buscarProdutoTipo(TipoProduto tipo) throws RemoteException, ProdutoInexistenteException {
		return this.gerenciadorProduto.buscarProdutoTipo(tipo);
	}
	
	@Override
	public synchronized Promocao buscarPromocao(long id) throws PromocaoInexistenteException, RemoteException{
		return this.gerenciadorPromocao.buscarPromocao(id);
	}

	@Override
	public synchronized Promocao buscarPromocao(String nome) throws RemoteException, PromocaoInexistenteException {
		return this.gerenciadorPromocao.buscarPromocao(nome);
	}
	
	@Override
	public synchronized Reserva buscaReserva(long id) throws ReservaInexistenteException, RemoteException{
		return this.gerenciadorReserva.buscarReserva(id);
	}

	@Override
	public synchronized Reserva buscarReserva(long sessao, Cliente cliente, Copia copia) throws RemoteException, UsuarioNaoLogadoException, ReservaInexistenteException, UsuarioSemPermissaoException {
		if(this.gerenciadorUsuario.estaLogado(sessao)){
			Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
			Reserva reserva = this.gerenciadorReserva.buscarReserva(cliente, copia);

			if(transacao instanceof Cliente){
				if(reserva.getCliente().equals(transacao)){
					return reserva;
				} else {
					throw new UsuarioSemPermissaoException(transacao);
				}
			} else if(transacao instanceof Dependente){
				if(reserva.getCliente().equals(((Dependente) transacao).getCliente())){
					return reserva;
				} else {
					throw new UsuarioSemPermissaoException(transacao);
				}
			} else {
				return reserva;
			}
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	@Override
	public synchronized IteratorBidirecional<Reserva> buscarReservas(long sessao, Copia copia) throws RemoteException, UsuarioNaoLogadoException, ReservaInexistenteException, UsuarioSemPermissaoException {
		if(this.gerenciadorUsuario.estaLogado(sessao)){
			Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
			if(transacao instanceof Administrador || transacao instanceof Funcionario){
				return this.gerenciadorReserva.buscarReservas(copia);
			} else {
				throw new UsuarioSemPermissaoException(transacao);
			}
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	@Override
	public synchronized IteratorBidirecional<Reserva> buscarReservas(long sessao, Cliente cliente) throws RemoteException, ReservaInexistenteException, UsuarioNaoLogadoException, UsuarioSemPermissaoException {
		if(this.gerenciadorUsuario.estaLogado(sessao)){
			Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
			if(transacao instanceof Administrador || transacao instanceof Funcionario){
				return this.gerenciadorReserva.buscarReservas(cliente);
			} else {
				throw new UsuarioSemPermissaoException(transacao);
			}
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	@Override
	public synchronized Usuario buscarUsuario(long sessao, long cpf) throws RemoteException, UsuarioNaoLogadoException, UsuarioInexistenteException, UsuarioSemPermissaoException {
		Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
		Usuario user = this.gerenciadorUsuario.buscarUsuario(cpf);

		if(transacao instanceof Funcionario || transacao instanceof Administrador){
			return user;
		} else if(user.equals(transacao)){
			return user;
		} else{
			throw new UsuarioSemPermissaoException(transacao);
		}
	}

	@Override
	public synchronized void cadastrarUsuario(long sessao, Usuario usuario) throws RemoteException, UsuarioSemPermissaoException, UsuarioExistenteException, UsuarioNaoLogadoException {
		this.gerenciadorUsuario.cadastrarUsuario(usuario, sessao);
	}

	@Override
	public synchronized void descadastrarUsuario(long sessao, long cpfUsuario, String senha) throws RemoteException, UsuarioInexistenteException, UsuarioSemPermissaoException, UsuarioNaoLogadoException, SenhaInvalidaException {
		Usuario user = this.gerenciadorUsuario.buscarUsuario(cpfUsuario);
		
		try {
			this.gerenciadorAvaliacaoProduto.removerAvaliacoes(user);
		} catch (AvaliacaoInexistenteException e) {
		}
		
		this.gerenciadorUsuario.descadastrarUsuario(user, sessao, senha);
	}

	@Override
	public synchronized void deslogar(long sessao) throws RemoteException, UsuarioNaoLogadoException {
		this.gerenciadorUsuario.deslogar(sessao);
	}

	@Override
	public synchronized IteratorBidirecional<Plano> getPlanos() throws RemoteException, PlanoInexistenteException {
		return this.gerenciadorPlano.getPlanos();
	}

	@Override
	public synchronized IteratorBidirecional<Produto> getProdutos() throws RemoteException, ProdutoInexistenteException {
		return this.gerenciadorProduto.getProdutos();
	}

	@Override
	public synchronized IteratorBidirecional<Promocao> getPromocoes() throws RemoteException, PromocaoInexistenteException {
		return this.gerenciadorPromocao.getPromocoes();
	}
	
	@Override
	public synchronized void inserirLocacao(long sessao, Locacao locacao) throws RemoteException, UsuarioNaoLogadoException, LocacaoExistenteException, LocacaoInvalidaException, UsuarioSemPermissaoException, CopiaLocadaNoPeriodoException, CopiaReservadaNoPeriodoException {
		Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);

		if(this.podeInserirLocacao(sessao, locacao)){
			try {
				Cliente user = (Cliente) this.gerenciadorUsuario.buscarUsuario(locacao.getCliente().getCpf());
				if(user.getLocacoesGratuitas() > 0){
					locacao.setGratuita(true);
				}
				
				this.locar(locacao);
				
				if(user.getLocacoesGratuitas() > 0){
					user.decrementarLocacoesGratuitas();
					this.gerenciadorUsuario.alterarUsuario(user);
				}
			} catch (UsuarioInexistenteException e) {
				// NAO vai acontecer
			} catch (UsuarioExistenteException e) {
				// NAO vai acontecer
			}
		} else {
			throw new UsuarioSemPermissaoException(transacao);
		}
	}
	
	protected boolean podeInserirLocacao(long sessao, Locacao locacao) {
		try{
			Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
			boolean retorno = false;
			
			if(transacao instanceof Administrador || transacao instanceof Funcionario){
				retorno = true;
			} else if(transacao instanceof Cliente){
				if(locacao.getCliente().getCpf() == transacao.getCpf()){
					retorno = true;
				} 
			} else {
				if(locacao.getCliente().getCpf() == ((Dependente) transacao).getCliente().getCpf()){
					retorno = true;
				}
			}
			
			return retorno;
		} catch(UsuarioNaoLogadoException e){
			return false;
		}
		
	}

	private synchronized void locar(Locacao locacao) throws CopiaReservadaNoPeriodoException, LocacaoExistenteException, LocacaoInvalidaException, CopiaLocadaNoPeriodoException{
		try {
			IteratorBidirecional<Reserva> reservas = this.gerenciadorReserva.buscarReservasInterceptaPeriodo(locacao.getCopia(), locacao.getPeriodo());

			Reserva reserva = reservas.next();

			if(reservas.hasNext()){
				throw new CopiaReservadaNoPeriodoException(locacao.getCopia(), locacao.getPeriodo());
			} else {
				if(reserva.getCliente().equals(locacao.getCliente())){
					this.removerReserva(reserva.getCliente(), reserva.getCopia());
					this.gerenciadorLocacao.inserirLocacao(locacao);
				}
			}
		} catch (ReservaInexistenteException e) {
			this.gerenciadorLocacao.inserirLocacao(locacao);
		}
	}
	
	public synchronized void removerReserva(Cliente cliente, Copia copia) throws ReservaInexistenteException {
		Filial filial = copia.getFilial();
		Reserva reserva = this.gerenciadorReserva.buscarReserva(cliente, copia);
		this.gerenciadorReserva.removerReserva(cliente, copia);
		
		if(filial != null && this.gerenciadorFilial.estaLogada(filial.getNome())){
			try {
				try {
					String ipConfig = String.format("rmi://%s:1099/",InterfaceGeral.ip);
					InterfaceFilial servico = (InterfaceFilial) Naming.lookup(ipConfig + "InterfaceFilial_" + this.gerarNome(filial.getNome()));
					
					new ThreadNotificar(servico, new Object[]{reserva}, "notificarRemocaoReserva").start();
				} catch (MalformedURLException e) {
					// NAO deve ocorrer
					e.printStackTrace();
				} catch (RemoteException e) {
					this.gerenciadorFilial.deslogarFilial(filial.getNome());
				} catch (NotBoundException e) {
					this.gerenciadorFilial.deslogarFilial(filial.getNome());
				}
			} catch (FilialNaoLogadaException e1) {
				// NAO deve ocorrer
				e1.printStackTrace();
			} catch(FilialInexistenteException e1){
				//NAO deve ocorrer
				e1.printStackTrace();
			}
		}
	}

	@Override
	public synchronized void inserirReserva(long sessao, Reserva reserva) throws RemoteException, UsuarioNaoLogadoException, ReservaExistenteException, UsuarioSemPermissaoException, CopiaLocadaNoPeriodoException, CopiaReservadaNoPeriodoException, FilialNaoLogadaException, FilialInexistenteException {
		Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
		Copia c = reserva.getCopia();
		InterfaceFilial servico = null;
		
		if(!(transacao instanceof Funcionario && ((Funcionario)transacao).getFilial().getId() == c.getFilial().getId())){
			String nomeFilial = null;
			
			try {
				String ipConfig = String.format("rmi://%s:1099/",InterfaceGeral.ip);
				nomeFilial = c.getFilial().getNome();
				servico = (InterfaceFilial) Naming.lookup(ipConfig + "InterfaceFilial_" + this.gerarNome(nomeFilial));
			} catch (MalformedURLException e) {
				// NAO deve ocorrer
				e.printStackTrace();
			} catch (RemoteException e) {
				this.gerenciadorFilial.deslogarFilial(nomeFilial);
				throw new FilialNaoLogadaException(nomeFilial);
			} catch (NotBoundException e) {
				this.gerenciadorFilial.deslogarFilial(nomeFilial);
				throw new FilialNaoLogadaException(nomeFilial);
			}
		}
		
		if(transacao instanceof Administrador || transacao instanceof Funcionario){
			this.reservar(reserva);
		} else if(transacao instanceof Cliente){
			if(reserva.getCliente().equals(transacao)){
				this.reservar(reserva);
			} else {
				throw new UsuarioSemPermissaoException(transacao);
			}
		} else {
			if(reserva.getCliente().equals(((Dependente) transacao).getCliente())){
				this.reservar(reserva);
			} else {
				throw new UsuarioSemPermissaoException(transacao);
			}
		}
		
		if(servico != null){
			new ThreadNotificar(servico, new Object[]{reserva}, "notificarReserva").start();
		}
	}

	protected String gerarNome(String nome){
		for(int i = 0; i < nome.length(); i++){
			if(nome.charAt(i) == ' '){
				nome = nome.substring(0, i) + nome.substring(i+1, nome.length());
			}
		}

		return nome;
	}

	private void reservar(Reserva reserva) throws CopiaLocadaNoPeriodoException, ReservaExistenteException, CopiaReservadaNoPeriodoException{
		try {
			this.gerenciadorLocacao.buscarLocacoesInterceptaPeriodo(reserva.getCopia(), reserva.getPeriodo());
			throw new CopiaLocadaNoPeriodoException(reserva.getCopia(), reserva.getPeriodo());
		} catch (LocacaoInexistenteException e) {
			this.gerenciadorReserva.inserirReserva(reserva);

			Timer t = new Timer();
			t.schedule(new TimerReserva(reserva, this), reserva.getPeriodo().getDataFim().getTime());
		}
	}

	@Override 
	public synchronized void alterarTelefonesCliente(long sessao, long cpf, Telefone[] telefones) throws RemoteException, UsuarioNaoLogadoException, UsuarioInexistenteException, UsuarioSemPermissaoException{
		this.gerenciadorUsuario.alterarTelefonesCliente(sessao, cpf, telefones);
	}

	@Override
	public synchronized long logarUsuario(long cpf, String senha) throws RemoteException, UsuarioLogadoException, UsuarioInexistenteException, SenhaInvalidaException {
		return this.gerenciadorUsuario.logar(cpf, senha);
	}

	@Override
	public synchronized void removerLocacao(long sessao, Copia copia) throws RemoteException, UsuarioNaoLogadoException, UsuarioSemPermissaoException, LocacaoInexistenteException {
		Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);

		if(transacao instanceof Administrador || transacao instanceof Funcionario){
			Locacao locacao = this.gerenciadorLocacao.buscarLocacao(copia);
			this.gerenciadorLocacao.removerLocacao(locacao);
		} else {
			throw new UsuarioSemPermissaoException(transacao);
		}
	}

	@Override
	public synchronized void removerReserva(long sessao, long cpf, int codigoCopia) throws RemoteException, ReservaInexistenteException, UsuarioNaoLogadoException, UsuarioSemPermissaoException, UsuarioInexistenteException, CopiaInexistenteException, ProdutoInexistenteException {
		Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
		Cliente cliente = (Cliente) this.gerenciadorUsuario.buscarUsuario(cpf);
		
		if(transacao instanceof Administrador || transacao instanceof Funcionario || transacao.equals(cliente)){
			Copia copia = this.gerenciadorProduto.buscarCopia(codigoCopia);
			this.removerReserva(cliente, copia);
		} else {
			throw new UsuarioSemPermissaoException(transacao);
		}
	}

	@Override
	public synchronized IteratorBidirecional<Locacao> buscarLocacoesInterceptaPeriodo(Copia copia, Periodo periodo) throws LocacaoInexistenteException, RemoteException {
		return this.gerenciadorLocacao.buscarLocacoesInterceptaPeriodo(copia, periodo);
	}

	@Override
	public synchronized IteratorBidirecional<Reserva> buscarReservasInterceptaPeriodo(long sessao, Copia copia, Periodo periodo) throws ReservaInexistenteException, UsuarioNaoLogadoException, UsuarioSemPermissaoException, RemoteException {
		if(this.gerenciadorUsuario.estaLogado(sessao)){
			Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);
			if(transacao instanceof Administrador || transacao instanceof Funcionario){
				return this.gerenciadorReserva.buscarReservasInterceptaPeriodo(copia, periodo);
			} else {
				throw new UsuarioSemPermissaoException(transacao);
			}
		} else {
			throw new UsuarioNaoLogadoException();
		}
	}

	@Override
	public synchronized double devolverProduto(long sessao, int codigoCopia, String filial) throws UsuarioNaoLogadoException, UsuarioSemPermissaoException, LocacaoInexistenteException, ProdutoInexistenteException, CopiaInexistenteException, FilialInvalidaException, RemoteException, FilialInexistenteException, UsuarioInexistenteException {
		Usuario transacao = this.gerenciadorUsuario.getUsuarioLogado(sessao);

		if(transacao instanceof Administrador || transacao instanceof Funcionario){
			Copia copia = this.gerenciadorProduto.buscarCopia(codigoCopia);
			Locacao locacao = this.gerenciadorLocacao.buscarLocacao(copia);
			this.gerenciadorLocacao.removerLocacao(locacao);
			
			Filial f = this.gerenciadorFilial.buscarFilial(filial);
			this.gerenciadorProduto.mudarFilialDeCopia(locacao.getCopia(), f);
			return locacao.getCusto();
		} else {
			throw new UsuarioSemPermissaoException(transacao);
		}

	}

	@Override
	public synchronized Copia buscarCopia(int codigoCopia) throws CopiaInexistenteException, RemoteException{
		return this.gerenciadorProduto.buscarCopia(codigoCopia);
	}
	
	@Override
	public Copia getCopiaFilial(String nomeProduto, String nomeFilial) throws RemoteException, CopiaInexistenteException, ProdutoInexistenteException{
		IteratorBidirecional<Copia> copias = this.gerenciadorProduto.buscarProduto(nomeProduto).getCopias();
		Copia c;
		
		while(copias.hasNext()){
			c = copias.next();
			if(!this.gerenciadorLocacao.existeLocacao(c) && c.getFilial().getNome().equals(nomeFilial)){
				return c;
			}
		}
		
		throw new CopiaInexistenteException();
	}
	
	public static void main(String[] args) {
		int port = 1099;
		
		try {
			LocateRegistry.createRegistry( port );
			Registry registry = LocateRegistry.getRegistry();

			if ( System.getSecurityManager() == null ){
				System.setSecurityManager(new RMISecurityManager());
			}

			String name = "InterfaceFilialBD";
			InterfaceFilialBD obj = new InterfaceFilialBDImp();

			registry.rebind(name, obj);
			System.out.println( " Interface Filial registrada. ");

			name = "InterfaceWebBD";
			InterfaceWebBD obj2 = new InterfaceWebBDImp();

			registry.rebind(name, obj2);
			System.out.println( " Interface Web registrada. ");
		} catch (RemoteException e) {
			e.printStackTrace();
		}
	}

}
